"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Merges multiple GTFS sets into a single stream for each GTFS file (stops.txt etc)
 */
class MergedGTFS {
    constructor(calendarStream, calendarDatesStream, tripsStream, stopTimesStream, routesStream, agencyStream, stopsStream, transfersStream, transferDistance) {
        this.calendarStream = calendarStream;
        this.calendarDatesStream = calendarDatesStream;
        this.tripsStream = tripsStream;
        this.stopTimesStream = stopTimesStream;
        this.routesStream = routesStream;
        this.agencyStream = agencyStream;
        this.stopsStream = stopsStream;
        this.transfersStream = transfersStream;
        this.transferDistance = transferDistance;
        this.calendarHashes = {};
        this.currentServiceId = 1;
        this.currentTripId = 1;
        this.stopLocations = {};
        this.parentStops = {};
        this.additionalTransfers = {};
        this.dummyCalendar = {
            monday: 0,
            tuesday: 0,
            wednesday: 0,
            thursday: 0,
            friday: 0,
            saturday: 0,
            sunday: 0,
            start_date: "19700101",
            end_date: "20991231",
        };
    }
    /**
     * Merge in the given GTFS data set and push the new items to the file streams.
     */
    merge(gtfs) {
        const serviceIdMap = {};
        this.writeStops(this.stopsStream, gtfs.stops);
        this.writeCalendars(gtfs.calendars, gtfs.calendarDates, serviceIdMap);
        this.writeTrips(gtfs.trips, gtfs.stopTimes, serviceIdMap);
        this.writeTransfers(this.transfersStream, gtfs.transfers);
        this.writeAll(this.routesStream, gtfs.routes);
        this.writeAll(this.agencyStream, gtfs.agencies);
    }
    writeCalendars(calendars, dateIndex, serviceIdMap) {
        for (const calendar of calendars) {
            const calendarDates = dateIndex[calendar.service_id] || [];
            this.writeCalendar(calendar, calendarDates, serviceIdMap);
        }
        // check for any calendar dates that have no calendar entry
        for (const serviceId of Object.keys(dateIndex)) {
            if (!serviceIdMap[serviceId]) {
                const calendarDates = dateIndex[serviceId];
                const dummyCalendar = Object.assign({ service_id: serviceId }, this.dummyCalendar);
                this.writeCalendar(dummyCalendar, calendarDates, serviceIdMap);
            }
        }
    }
    writeCalendar(calendar, calendarDates, serviceIdMap) {
        const hash = getCalendarHash(calendar, calendarDates);
        if (!this.calendarHashes[hash]) {
            const serviceId = this.currentServiceId++;
            this.calendarHashes[hash] = serviceId;
            for (const calendarDay of calendarDates) {
                calendarDay.service_id = serviceId;
                this.calendarDatesStream.write(calendarDay);
            }
            serviceIdMap[calendar.service_id] = serviceId;
            calendar.service_id = serviceId;
            this.calendarStream.write(calendar);
        }
        else {
            serviceIdMap[calendar.service_id] = this.calendarHashes[hash];
        }
    }
    writeTrips(trips, stopTimes, serviceIdMap) {
        const tripIdMap = {};
        for (const trip of trips) {
            tripIdMap[trip.trip_id] = this.currentTripId++;
            trip.trip_id = tripIdMap[trip.trip_id];
            trip.service_id = serviceIdMap[trip.service_id];
            this.tripsStream.write(trip);
        }
        for (const stopTime of stopTimes) {
            stopTime.trip_id = tripIdMap[stopTime.trip_id];
            stopTime.stop_id = this.parentStops[stopTime.stop_id] || stopTime.stop_id;
            this.stopTimesStream.write(stopTime);
        }
    }
    writeAll(stream, items) {
        for (const item of items) {
            stream.write(item);
        }
    }
    writeTransfers(stream, transfers) {
        for (const transfer of transfers) {
            stream.write(transfer);
            this.additionalTransfers[transfer.from_stop_id + "," + transfer.to_stop_id] = true;
        }
    }
    writeStops(stream, stops) {
        for (const stop of stops) {
            if (stop.parent_station) {
                this.parentStops[stop.stop_id] = stop.parent_station;
            }
            else {
                stream.write(stop);
                if (!this.stopLocations[stop.stop_id]) {
                    this.addNearbyStops(stop);
                }
            }
        }
    }
    /**
     * Search any stops we've seen to see if we can walk there
     */
    addNearbyStops(stop) {
        const aLon = Number(stop.stop_lon);
        const aLat = Number(stop.stop_lat);
        for (const stopId in this.stopLocations) {
            const [bLon, bLat] = this.stopLocations[stopId];
            const key = stop.stop_id + "," + stopId;
            const reverseKey = stopId + "," + stop.stop_id;
            if (bLat === 0.0 && bLon === 0.0) {
                continue;
            }
            if (!this.additionalTransfers[key] || !this.additionalTransfers[reverseKey]) {
                const distance = this.getDistance(aLon, aLat, bLon, bLat);
                if (distance < this.transferDistance) {
                    const t = Math.max(60, Math.round(distance * 72000));
                    const transfer = { from_stop_id: stop.stop_id, to_stop_id: stopId, transfer_type: 2, min_transfer_time: t };
                    const reverse = { from_stop_id: stopId, to_stop_id: stop.stop_id, transfer_type: 2, min_transfer_time: t };
                    this.additionalTransfers[key] = this.additionalTransfers[key] || transfer;
                    this.additionalTransfers[reverseKey] = this.additionalTransfers[reverseKey] || reverse;
                }
            }
        }
        this.stopLocations[stop.stop_id] = [aLon, aLat];
    }
    /**
     * Note this method of calculating distances between stations is flawed and only used as a rough guide.
     */
    getDistance(aLon, aLat, bLon, bLat) {
        return Math.abs(bLat - aLat) + Math.abs(bLon - aLon);
    }
    async end() {
        for (const transfer of Object.values(this.additionalTransfers)) {
            if (transfer !== true) {
                this.transfersStream.write(transfer);
            }
        }
        await Promise.all([
            this.calendarDatesStream.end(),
            this.calendarStream.end(),
            this.tripsStream.end(),
            this.stopTimesStream.end(),
            this.routesStream.end(),
            this.agencyStream.end(),
            this.stopsStream.end(),
            this.transfersStream.end()
        ]);
    }
}
exports.MergedGTFS = MergedGTFS;
function getCalendarHash(calendar, calendarDates) {
    const { service_id, ...rest } = calendar;
    const days = calendarDates.map(d => d.date + "_" + d.exception_type).join(":");
    const fields = Object.values({ days, ...rest });
    return fields.join();
}
